from __future__ import annotations
import numpy as np
import os
from enum import Enum
from .enum import Residue, RES_ABBREV, RibonucleicAcid, Element
from .reduction import Reduction, _Reduction, REDUCTIONS
from typing import Generator
import torch
from copy import copy
from ._c import _load


UNKNOWN = "UNKNOWN"


class Scale(Enum):
    RESIDUE  = 0
    CHAIN    = 1
    MOLECULE = 2


class Molecule(Enum):
    PROTEIN = 0
    RNA     = 1
    DNA     = 2
    WATER   = 3
    ION     = 4


def _molecule_type(val: int) -> Molecule:

    if val == 0:
        return Molecule.PROTEIN
    if val == 1:
        return Molecule.RNA
    if val == 2:
        return Molecule.DNA
    if val == 3:
        return Molecule.OTHER
    if val == 4:
        return Molecule.MISSING


def _all_equal(*values: int) -> bool:

    return len(set(values)) == 1


class Polymer:

    def __init__(
        self: Polymer,
        coordinates: torch.Tensor,
        atoms: torch.Tensor,
        elements: torch.Tensor,
        sequence: torch.Tensor,
        sizes: dict[Scale, torch.Tensor],
        id: str,
        names: list[str],
        lengths: torch.Tensor,
        nonpoly: int = 0,
    ) -> None:

        self._id = id
        self.names = names
        self.nonpoly = nonpoly

        if not _all_equal(
            coordinates.size(0),
            atoms.size(0),
            elements.size(0),
        ):
            raise ValueError(f"The coordinate, atom, and element tensors must be equal in size for PDB {self.id()}.")

        _res_count = sizes[Scale.RESIDUE].sum().item()
        _chn_count = sizes[Scale.CHAIN].sum().item()
        _mol_count = sizes[Scale.MOLECULE].sum().item()

        if not _all_equal(_res_count + nonpoly, _chn_count, _mol_count):
            raise ValueError(f"The residue ({_res_count} + {nonpoly}), chain ({_chn_count}), and molecule ({_mol_count}) atom counts do not agree for PDB {self.id()}.")

        self.coordinates = coordinates
        self.atoms = atoms
        self.elements = elements
        self.sequence = sequence
        self._sizes = sizes
        self.lengths = lengths

    def id(self: Polymer) -> str:
        """
        Get the PDB ID of the molecule.
        """

        if not self._id:
            return UNKNOWN
        return self._id

    def chain_id(self: Polymer, ix: int | None = None) -> str | list[str]:
        """
        Get the PDB + chain ID of each (or the specified) chain.
        """

        if ix is not None:
            return self.id() + '_' + self.names[ix]

        return [self.id() + '_' + name for name in self.names]

    def size(self: Polymer, scale: Scale | None = None) -> int:
        """
        Get the number of objects of the specified scale. Defaults to the
        number of atoms.
        """

        if scale is None:
            return self.coordinates.size(0)
        else:
            return self._sizes[scale].size(0)

    def type(self: Polymer) -> torch.Tensor:
        """
        Get the type of molecule (protein, RNA, DNA) of each chain.
        """

        types = torch.zeros(self.size(Scale.CHAIN), dtype=torch.long)
        atoms, _ = self.rreduce(self.sequence, Scale.CHAIN, Reduction.MAX)

        types[atoms < 5] = Molecule.RNA.value

        return types

    def reduce(
        self: Polymer,
        features: torch.Tensor,
        scale: Scale,
        rtype: Reduction = Reduction.MEAN,
    ) -> _Reduction:
        """
        Reduce the input values within each copy of the given object.
        MIN and MAX reductions return the indices too. A COLLATE
        reduction instead returns a list of tensors containing the
        values aligning with each specific index.
        """

        ATOM_DIM = 0

        count = self.size(scale)
        sizes = self._sizes[scale]
        ix = torch.arange(count).repeat_interleave(sizes)

        return REDUCTIONS[rtype](
            features, ix,
            dim=ATOM_DIM,
            dim_size=count,
        )

    def rreduce(
        self: Polymer,
        features: torch.Tensor,
        scale: Scale,
        rtype: Reduction = Reduction.MEAN,
    ) -> _Reduction:
        """
        Similar to reduce(), but with one feature per residue rather than per
        atom.
        """

        ATOM_DIM = 0

        count = self.size(scale)
        ix = torch.arange(count).repeat_interleave(self.lengths)

        return REDUCTIONS[rtype](
            features, ix,
            dim=ATOM_DIM,
            dim_size=count,
        )

    def expand(
        self: Polymer,
        features: torch.Tensor,
        scale: Scale,
    ) -> torch.Tensor:
        """
        Expand the features so that there is one per atom, rather than
        one per [scale].
        """

        return features.repeat_interleave(self._sizes[scale], dim=0)

    def center(
        self: Polymer,
        scale: Scale = Scale.MOLECULE,
    ) -> tuple[Polymer, torch.Tensor]:
        """
        Center the coordinates of each of the chosen object in the molecule.
        Return the centers of each [scale] well.
        """

        # Get the coordinate means

        means = self.reduce(self.coordinates, scale)

        # Expand the means and subtract

        expanded = self.expand(means, scale)
        coordinates = self.coordinates - expanded

        # Create a new Polymer with the centered coordinates

        centered = copy(self)
        centered.coordinates = coordinates

        return centered, means

    def _pc(
        self: Polymer,
        scale: Scale,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Compute the principal components of all atoms for each instance
        of the given property.

        Since principal components are only defined up to sign,
        the outputs of this function may be unstable with respect to
        the coordinates. See the .align method for getting stable
        principal components.
        """

        # Get the coordinate covariance matrices at the given scale

        cov = self.coordinates[:, None, :] * self.coordinates[:, :, None]
        cov = self.reduce(cov, scale)

        # Compute the eigenvectors of the covariance matrices

        return torch.linalg.eigh(cov)

    def align(
        self: Polymer,
        scale: Scale,
    ) -> tuple[Polymer, torch.Tensor]:
        """
        Align the objects such that their covariance matrix is diagonal.
        Also return the principal components used to align the objects.
        """

        # Center and get the principal components

        aligned, _ = self.center(scale)
        _, Q = aligned._pc(scale)

        # Expand the components and use them to rotate the coordinates

        Q_exp = aligned.expand(Q, scale)
        aligned.coordinates = (
            Q_exp @ aligned.coordinates[..., None]
        ).squeeze()

        # To ensure stability and uniqueness, we modify the coordinates
        # and Q so that the largest two third moments are positive. The
        # third is chosen such that the system is right-handed.

        signs = aligned.moment(3, scale).sign()

        # The smallest eigenvalue is always the first

        signs[:, 0] = signs[:, 1] * signs[:, 2] * torch.linalg.det(Q)
        signs_exp = aligned.expand(signs, scale)

        # Modify the signs and return

        aligned.coordinates = aligned.coordinates * signs_exp
        Q = Q * signs[..., None]

        return aligned, Q

    def count(
        self: Polymer,
        mask: torch.Tensor,
        scale: Scale,
    ) -> torch.Tensor:
        """
        Count the nonzero elements per each [scale].
        """

        return self.reduce(mask.long(), scale, Reduction.SUM)

    def __getitem__(
        self: Polymer,
        mask: torch.Tensor,
    ) -> Polymer:
        """
        Return a polymer containing only the selected atoms.
        """

        coordinates = self.coordinates[mask]
        atoms = self.atoms[mask]
        elements = self.elements[mask]

        chain_sizes = self.count(mask, Scale.CHAIN)
        res_sizes = self.count(mask, Scale.RESIDUE)
        mol_sizes = self.count(mask, Scale.MOLECULE)

        _residues = torch.zeros(self.size(Scale.CHAIN), dtype=torch.bool)
        _residues[chain_sizes > 0] = True
        residues = _residues.repeat_interleave(self.lengths, dim=0)

        lengths = self.lengths[chain_sizes > 0]

        sizes = {
            Scale.RESIDUE: res_sizes[residues],
            Scale.CHAIN: chain_sizes[chain_sizes > 0],
            Scale.MOLECULE: mol_sizes,
        }

        sequence = self.sequence[residues]

        names = [
            self.names[ix] for ix in range(len(self.names))
            if chain_sizes[ix] > 0
        ]

        return Polymer(
            coordinates,
            atoms,
            elements,
            sequence,
            sizes,
            self._id,
            names,
            lengths,
        )

    def mask(
        self: Polymer,
        mask: torch.Tensor | int,
        scale: Scale,
    ) -> torch.Tensor:
        """
        Create a boolean mask to select only the indicated [scale]s.
        """

        counts = self.size(scale)
        objects = torch.zeros(counts, dtype=torch.bool)
        objects[mask] = True

        return self.expand(objects, scale)

    def select(
        self: Polymer,
        ix: torch.Tensor | int,
        scale: Scale,
    ) -> Polymer:
        """
        Create a new Polymer containing only the selected objects.
        """

        return self[self.mask(ix, scale)]

    def get_by_name(
        self: Polymer,
        name: torch.Tensor | int,
    ) -> Polymer:
        """
        Similar to select(), but for atom name rather than scale.
        """

        ix = (self.atoms[:, None] == name).any(1)
        return self[ix]

    def subset(
        self: Polymer,
        mol: Molecule,
    ) -> Polymer:
        """
        Similar to select(), but for molecule type rather than scale.
        """

        ix = (self.type() == mol.value).nonzero().squeeze(-1)
        return self.select(ix, Scale.CHAIN)

    def chains(
        self: Polymer,
        mol: Molecule | None = None
    ) -> Generator[Polymer]:
        """
        Yield chains from the polymer. Only yeild those of the given type if
        specified.
        """

        for ix in range(self.size(Scale.CHAIN)):
            chain = self.select(ix, Scale.CHAIN)
            if mol is None or chain.istype(mol):
                yield chain

    def missing(
        self: Polymer,
        scale: Scale = Scale.RESIDUE,
    ) -> torch.Tensor:
        """
        The indices of missing objects in the polymer. Defaults to missing
        residues.
        """

        return (self._sizes[scale] == 0).nonzero().squeeze(-1)

    def moment(
        self: Polymer,
        n: int,
        scale :Scale,
    ) -> torch.Tensor:
        """
        Return the n-th (uncentered) moment of the coordinates of the
        polymer at the given scale.
        """

        return self.reduce(self.coordinates ** n, scale)

    def __repr__(self: Polymer) -> str:

        out = ""

        out += f"PDB {self.id()} with {self.size()} atoms.\n"
        out += "───────────────────────────────────────\n"
        out += " " * len(f"{self.size(Scale.CHAIN)}") + "  "
        out += "Type     # Res  # Atom"
        out += "\n"

        types = self.type()

        for ix in range(self.size(Scale.CHAIN)):

            mol = _molecule_type(types[ix])

            _chain = f"{self.names[ix]}"
            _type = f"{mol.name}"
            _residues = f"{self.lengths[ix]}"
            _atoms = f"{self._sizes[Scale.CHAIN][ix]}"

            out += _chain + " " * (2 - len(_chain)) + "  "
            out += _type + " " * (9 - len(_type))
            out += _residues + " " * (7 - len(_residues))
            out += _atoms
            out += "\n"

        return out

    def str(self: Polymer) -> str:
        """
        Get the sequence of residues as a string.
        """

        DICT = lambda x: RES_ABBREV[Residue.revdict().get(x, 'N')]
        return "".join([DICT(ix.item()) for ix in self.sequence])

    def istype(
        self: Polymer,
        mol: Molecule,
    ) -> bool:
        """
        Check if the chain is the specified molecule type. Only works for
        single chains.
        """

        _type = self.type()
        if _type.size(0) != 1:
            return False
        return _type[0].item() == mol.value

    def write(self: Polymer, filename: str) -> None:

        with open(filename, 'w') as file:

            for chain in self.chains():

                atom = 0
                for residue in range(chain.size(Scale.RESIDUE)):

                    residue_name = self.str()[residue]

                    for _ in range(chain._sizes[Scale.RESIDUE][residue]):

                        element = Element.revdict()[chain.elements[atom].item()]
                        try:
                            atom_name = RibonucleicAcid.revdict()[chain.atoms[atom].item()].replace('p', '\'')
                        except Exception:
                            breakpoint()
                        file.write("ATOM  {:5d} {:4s} {:3s} {:1s}{:4d}    {:8.3f}{:8.3f}{:8.3f}  1.00  0.00           {:2s}\n".format(
                            atom + 1,
                            atom_name,
                            residue_name,
                            chain.names[0],
                            residue + 1,
                            chain.coordinates[atom][0],
                            chain.coordinates[atom][1],
                            chain.coordinates[atom][2],
                            element,
                        ))

                        atom += 1


def load(file: str) -> Polymer:

    if not os.path.isfile(file):
        raise OSError(f"The file \"{file}\" does not exist.")

    id, coordinates, atoms, elements, residues, atoms_per_res, atoms_per_chain, res_per_chain, chain_names, nonpoly = _load(file)

    mol_sizes = torch.tensor([len(coordinates)], dtype=torch.long).numpy()

    sizes = {
        Scale.RESIDUE: atoms_per_res,
        Scale.CHAIN: atoms_per_chain,
        Scale.MOLECULE: mol_sizes,
    }

    return Polymer(
        torch.from_numpy(coordinates),
        torch.from_numpy(atoms).long(),
        torch.from_numpy(elements).long(),
        torch.from_numpy(residues).long(),
        {key: torch.from_numpy(value).long() for key, value in sizes.items()},
        id,
        chain_names,
        torch.from_numpy(res_per_chain),
        nonpoly,
    )



from __future__ import annotations
from enum import Enum
import torch
import itertools


class PairEnum(list):
    """
    Store a set of pairs of atom enums, as well as their integer
    representation in a tensor.
    """

    def __init__(
        self: PairEnum,
        bonds: list[tuple[Enum, Enum]]
    ) -> None:

        super().__init__(bonds)

    def __add__(
        self: PairEnum,
        other: list,
    ) -> PairEnum:

        return self.__class__(super().__add__(other))

    def indices(
        self: PairEnum,
    ) -> torch.Tensor:

        return  torch.tensor(
            [[atom1.value, atom2.value]
             for atom1, atom2 in self]
        )

    def pairwise(
        self: PairEnum,
    ) -> torch.Tensor:

        n = self.indices().max() + 1
        table = torch.ones(n, n, dtype=torch.long) * -1

        ix = 0
        for x, y in self:
            table[x.value, y.value] = ix
            table[y.value, x.value] = ix
            ix += 1

        return table


class IndexEnum(Enum):
    """
    An enum with the ability to collect all enum values into a tensor.
    """

    @classmethod
    def index(
        cls: type[IndexEnum],
    ) -> torch.Tensor:
        """
        Return a tensor of all the indices in the enum.
        """

        return torch.Tensor([
            atom.value for atom in cls
        ]).long()

    @classmethod
    def list(
        cls: type[IndexEnum],
        modifier: str = ''
    ) -> list[str]:
        """
        Return the names in the enum as a list.
        """

        return [
            modifier + field.name
            for field in cls
        ]

    @classmethod
    def dict(
        cls: type[IndexEnum],
        modifier: str = ''
    ) -> dict[str, int]:
        """
        Return the enum as a dict.
        """

        return {
            modifier + field.name: field.value
            for field in cls
        }

    @classmethod
    def revdict(
        cls: type[IndexEnum],
        modifier: str = ''
    ) -> dict[int, str]:
        """
        Return the enum as a reversed dict.
        """

        return {
            field.value: modifier + field.name
            for field in cls
        }

    @classmethod
    def pairs(
        cls: type[IndexEnum],
    ) -> PairEnum:
        """
        Return all pairs of objects, without repeats.
        """

        pairs = []
        for x, y in itertools.product(cls, cls):
            if (x, y) not in pairs and (y, x) not in pairs:
                pairs.append((x, y))

        return PairEnum(pairs)


class Residue(IndexEnum):
    A   = 0
    DA  = 0
    C   = 1
    DC  = 1
    G   = 2
    DG  = 2
    U   = 3
    DU  = 3
    T   = 4
    DT  = 4
    ALA = 5
    CYS = 6
    ASP = 7
    GLU = 8
    PHE = 9
    GLY = 10
    HIS = 11
    ILE = 12
    LYS = 13
    LEU = 14
    MET = 15
    ASN = 16
    PRO = 17
    GLN = 18
    ARG = 19
    SER = 20
    THR = 21
    VAL = 22
    TRP = 23
    TYR = 24
    # Water
    HOH = 25
    # Ions
    MG  = 26
    CS  = 27

RES_ABBREV = {
    'A': 'a',
    'C': 'c',
    'G': 'g',
    'U': 'u',
    'T': 't',
    'N' : 'n',
    'ALA': 'A',
    'CYS': 'C',
    'ASP': 'D',
    'GLU': 'E',
    'PHE': 'F',
    'GLY': 'G',
    'HIS': 'H',
    'ILE': 'I',
    'LYS': 'K',
    'LEU': 'L',
    'MET': 'M',
    'ASN': 'N',
    'PRO': 'P',
    'GLN': 'Q',
    'ARG': 'R',
    'SER': 'S',
    'THR': 'T',
    'VAL': 'V',
    'TRP': 'W',
    'TYR': 'Y'
}


class Adenosine(IndexEnum):
    OP3  = 1
    P    = 2
    OP1  = 3
    OP2  = 4
    O5p  = 5
    C5p  = 6
    C4p  = 7
    O4p  = 8
    C3p  = 9
    O3p  = 10
    C2p  = 11
    O2p  = 12
    C1p  = 13
    N9   = 14
    C8   = 15
    N7   = 16
    C5   = 17
    C6   = 18
    N6   = 19
    N1   = 20
    C2   = 21
    N3   = 22
    C4   = 23
    HOP3 = 24
    HOP2 = 25
    H5p  = 26
    H5pp = 27
    H4p  = 28
    H3p  = 29
    HO3p  = 30
    H2p  = 31
    HO2p  = 32
    HO5p = 148
    H1p  = 33
    H8   = 34
    H61  = 35
    H62  = 36
    H2   = 37

class Cytosine(IndexEnum):
    OP3  = 38
    P    = 39
    OP1  = 40
    OP2  = 41
    O5p  = 42
    C5p  = 43
    C4p  = 44
    O4p  = 45
    C3p  = 46
    O3p  = 47
    C2p  = 48
    O2p  = 49
    C1p  = 50
    N1   = 51
    C2   = 52
    O2   = 53
    N3   = 54
    C4   = 55
    N4   = 56
    C5   = 57
    C6   = 58
    HOP3 = 59
    HOP2 = 60
    H5p  = 61
    H5pp = 62
    H4p  = 63
    H3p  = 64
    HO3p  = 65
    H2p  = 66
    HO2p  = 67
    HO5p = 145
    H1p  = 68
    H41  = 69
    H42  = 70
    H5   = 71
    H6   = 72

class Guanosine(IndexEnum):
    OP3  = 73
    P    = 74
    OP1  = 75
    OP2  = 76
    O5p  = 77
    C5p  = 78
    C4p  = 79
    O4p  = 80
    C3p  = 81
    O3p  = 82
    C2p  = 83
    O2p  = 84
    C1p  = 85
    N9   = 86
    C8   = 87
    N7   = 88
    C5   = 89
    C6   = 90
    O6   = 91
    N1   = 92
    C2   = 93
    N2   = 94
    N3   = 95
    C4   = 96
    HOP3 = 97
    HOP2 = 98
    H5pp = 99
    H5p  = 100
    H4p  = 101
    H3p  = 102
    HO3p  = 103
    H2p  = 104
    HO2p  = 105
    HO5p = 146
    H1p  = 106
    H8   = 107
    H1   = 108
    H21  = 109
    H22  = 110

class Uridine(IndexEnum):
    OP3  = 111
    P    = 112
    OP1  = 113
    OP2  = 114
    O5p  = 115
    C5p  = 116
    C4p  = 117
    O4p  = 118
    C3p  = 119
    O3p  = 120
    C2p  = 121
    O2p  = 122
    C1p  = 123
    N1   = 124
    C2   = 125
    O2   = 126
    N3   = 127
    C4   = 128
    O4   = 129
    C5   = 130
    C6   = 131
    HOP3 = 132
    HOP2 = 133
    H5p  = 134
    H5pp = 135
    H4p  = 136
    H3p  = 137
    HO3p  = 138
    H2p  = 139
    HO2p  = 140
    HO5p = 147
    H1p  = 141
    H3   = 142
    H5   = 143
    H6   = 144

RibonucleicAcid = IndexEnum("RibonucleicAcid", Adenosine.dict() | Cytosine.dict() | Guanosine.dict() | Uridine.dict())

_A = {"A_" + key: value for key, value in Adenosine.dict().items() if 'p' in key or "P" in key}
_C = {"C_" + key: value for key, value in Cytosine.dict().items() if 'p' in key or "P" in key}
_G = {"G_" + key: value for key, value in Guanosine.dict().items() if 'p' in key or "P" in key}
_U = {"U_" + key: value for key, value in Uridine.dict().items() if 'p' in key or "P" in key}

Backbone = IndexEnum("Backbone", _A | _C | _G | _U)

_A = {"A_" + key: value for key, value in Adenosine.dict().items() if 'p' not in key and "P" not in key}
_C = {"C_" + key: value for key, value in Cytosine.dict().items() if 'p' not in key and "P" not in key}
_G = {"G_" + key: value for key, value in Guanosine.dict().items() if 'p' not in key and "P" not in key}
_U = {"U_" + key: value for key, value in Uridine.dict().items() if 'p' not in key and "P" not in key}

Nucleobase = IndexEnum("Nucleobase", _A | _C | _G | _U)

_A = {"A_" + key: value for key, value in Adenosine.dict().items() if "P" in key}
_C = {"C_" + key: value for key, value in Cytosine.dict().items() if "P" in key}
_G = {"G_" + key: value for key, value in Guanosine.dict().items() if "P" in key}
_U = {"U_" + key: value for key, value in Uridine.dict().items() if "P" in key}

Phosphate = IndexEnum("Phosphate", _A | _C | _G | _U)



COARSE = torch.tensor([
    # Adenosine.C1p.value,
    # Cytosine.C1p.value,
    # Guanosine.C1p.value,
    # Uridine.C1p.value,
    Adenosine.N1.value,
    Cytosine.N3.value,
    Guanosine.N1.value,
    Uridine.N3.value,
])



class Element(IndexEnum):
    H = 1
    C = 6
    N = 7
    O = 8
    P = 15
    S = 16



from enum import Enum
from typing import Union
import torch
from torch_scatter import (
    scatter_sum as t_scatter_sum,
    scatter_mean as t_scatter_mean,
    scatter_max as t_scatter_max,
    scatter_min as t_scatter_min,
)


class Reduction(Enum):
    NONE = 0
    COLLATE = 1
    MEAN = 2
    SUM = 3
    MIN = 4
    MAX = 5


def t_scatter_collate(
    features: torch.Tensor,
    indices: torch.Tensor,
    dim: int,
    dim_size: int,
) -> list[torch.Tensor]:
    """
    Return a list of tensors, containing all values corresponding to each value
    of the index.
    """

    return [
        features[indices == ix]
        for ix in range(indices.max() + 1)
    ]


REDUCTIONS = {
    Reduction.NONE: lambda features, indices, dim, dim_size: features,
    Reduction.COLLATE: t_scatter_collate,
    Reduction.MEAN: t_scatter_mean,
    Reduction.SUM: t_scatter_sum,
    Reduction.MIN: t_scatter_min,
    Reduction.MAX: t_scatter_max,
}


_Reduction = Union[
    torch.Tensor,
    tuple[torch.Tensor, torch.LongTensor],
    list[torch.Tensor],
]



__version__ = "0.1.4"

from .main import Polymer, Scale, Molecule, Reduction, load

RESIDUE = Scale.RESIDUE
CHAIN = Scale.CHAIN
MOLECULE = Scale.MOLECULE

PROTEIN = Molecule.PROTEIN
RNA = Molecule.RNA
DNA = Molecule.DNA



from setuptools import setup, Extension
import os
import re
import numpy

NAME = 'ciffy'


def _version() -> str:
    with open(os.path.join(os.path.dirname(__file__), NAME, '__init__.py')) as f:
        content = f.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Cannot find version information")


VERSION = _version()
LICENSE = 'CC BY-NC 4.0'
AUTHOR = 'Hamish M. Blair'
EMAIL = 'hmblair@stanford.edu'
URL = 'https://github.com/hmblair/ciffy'

EXT = "_c"
SOURCES = [
    'ciffy/src/_c.c',
    'ciffy/src/io.c',
    'ciffy/src/py.c',
    'ciffy/src/cif.c',
]
module = Extension(
    name=f"{NAME}.{EXT}",
    sources=SOURCES,
    include_dirs=[numpy.get_include(), f'{NAME}/src'],
    extra_compile_args=['-O3'],
)

setup(
    name=NAME,
    version=VERSION,
    packages=[NAME],
    ext_modules=[module],
    install_requires=['numpy'],
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    license=LICENSE,
)
